@@ -2,9 +2,10 @@ use std::{collections::HashMap, convert::TryFrom, fmt, future::Future, pin::Pin,
 
 use anyhow::{anyhow, bail, Context, Result};
 use async_trait::async_trait;
-use colibri::ColibriMessage;
+use colibri::{ColibriMessage, JsonMessage};
 use futures::stream::StreamExt;
 use gstreamer::prelude::{ElementExt, ElementExtManual, GstBinExt};
+use jitsi_xmpp_parsers::jingle::{Action, Jingle};
 use maplit::hashmap;
 use once_cell::sync::Lazy;
 use serde::Serialize;
@@ -14,16 +15,17 @@ use tracing::{debug, error, info, trace, warn};
 use uuid::Uuid;
 pub use xmpp_parsers::disco::Feature;
 use xmpp_parsers::{
-  disco::{DiscoInfoQuery, DiscoInfoResult},
+  caps::{self, Caps},
+  disco::{DiscoInfoQuery, DiscoInfoResult, Identity},
   ecaps2::{self, ECaps2},
-  hashes::Algo,
+  hashes::{Algo, Hash},
   iq::{Iq, IqType},
-  jingle::{Action, Jingle},
   message::{Message, MessageType},
-  muc::{Muc, MucUser},
+  muc::{user::Status as MucStatus, Muc, MucUser},
   nick::Nick,
   ns,
   presence::{self, Presence},
+  stanza_error::{DefinedCondition, ErrorType, StanzaError},
   BareJid, Element, FullJid, Jid,
 };
 
@@ -36,8 +38,13 @@ use crate::{
   xmpp::{self, connection::Connection},
 };
 
-static DISCO_INFO: Lazy<DiscoInfoResult> = Lazy::new(|| {
-  let mut features = vec![
+const DISCO_NODE: &str = "https://github.com/avstack/gst-meet";
+
+static DISCO_INFO: Lazy<DiscoInfoResult> = Lazy::new(|| DiscoInfoResult {
+  node: None,
+  identities: vec![Identity::new("client", "bot", "en", "gst-meet")],
+  features: vec![
+    Feature::new(ns::DISCO_INFO),
     Feature::new(ns::JINGLE_RTP_AUDIO),
     Feature::new(ns::JINGLE_RTP_VIDEO),
     Feature::new(ns::JINGLE_ICE_UDP),
@@ -45,25 +52,14 @@ static DISCO_INFO: Lazy<DiscoInfoResult> = Lazy::new(|| {
     Feature::new("urn:ietf:rfc:5888"), // BUNDLE
     Feature::new("urn:ietf:rfc:5761"), // RTCP-MUX
     Feature::new("urn:ietf:rfc:4588"), // RTX
-  ];
-  let gst_version = gstreamer::version();
-  if gst_version.0 >= 1 && gst_version.1 >= 19 {
-    // RTP header extensions are supported on GStreamer 1.19+
-    features.push(Feature::new("http://jitsi.org/tcc"));
-  }
-  else {
-    warn!("Upgrade GStreamer to 1.19 or later to enable RTP header extensions");
-  }
-  // Not supported yet:
-  // Feature::new("http://jitsi.org/opus-red")
-  DiscoInfoResult {
-    node: None,
-    identities: vec![],
-    features,
-    extensions: vec![],
-  }
+    Feature::new("http://jitsi.org/tcc"),
+  ],
+  extensions: vec![],
 });
 
+static COMPUTED_CAPS_HASH: Lazy<Hash> =
+  Lazy::new(|| caps::hash_caps(&caps::compute_disco(&DISCO_INFO), Algo::Sha_1).unwrap());
+
 #[derive(Debug, Clone, Copy)]
 enum JitsiConferenceState {
   Discovering,
@@ -74,11 +70,15 @@ enum JitsiConferenceState {
 #[derive(Debug, Clone)]
 pub struct JitsiConferenceConfig {
   pub muc: BareJid,
-  pub focus: FullJid,
+  pub focus: Jid,
   pub nick: String,
-  pub region: String,
+  pub region: Option<String>,
   pub video_codec: String,
   pub extra_muc_features: Vec<String>,
+  pub start_bitrate: u32,
+  pub stereo: bool,
+  #[cfg(feature = "log-rtp")]
+  pub log_rtp: bool,
 }
 
 #[derive(Clone)]
@@ -90,6 +90,7 @@ pub struct JitsiConference {
   pub(crate) external_services: Vec<xmpp::extdisco::Service>,
   pub(crate) jingle_session: Arc<Mutex<Option<JingleSession>>>,
   pub(crate) inner: Arc<Mutex<JitsiConferenceInner>>,
+  pub(crate) tls_insecure: bool,
 }
 
 impl fmt::Debug for JitsiConference {
@@ -119,6 +120,7 @@ pub(crate) struct JitsiConferenceInner {
     Option<Arc<dyn (Fn(JitsiConference, Participant) -> BoxedResultFuture) + Send + Sync>>,
   on_colibri_message:
     Option<Arc<dyn (Fn(JitsiConference, ColibriMessage) -> BoxedResultFuture) + Send + Sync>>,
+  presence: Vec<Element>,
   state: JitsiConferenceState,
   connected_tx: Option<oneshot::Sender<()>>,
 }
@@ -142,10 +144,8 @@ impl JitsiConference {
       machine_uid: Uuid::new_v4().to_string(),
       room: config.muc.to_string(),
       properties: hashmap! {
-        // Disable voice processing
-        // TODO put this in config
-        "stereo".to_string() => "true".to_string(),
-        "startBitrate".to_string() => "800".to_string(),
+        "stereo".to_string() => config.stereo.to_string(),
+        "startBitrate".to_string() => config.start_bitrate.to_string(),
       },
     };
 
@@ -153,6 +153,46 @@ impl JitsiConference {
 
     let focus = config.focus.clone();
 
+    let ecaps2_hash = ecaps2::hash_ecaps2(&ecaps2::compute_disco(&DISCO_INFO)?, Algo::Sha_256)?;
+    let mut presence = vec![
+      Muc::new().into(),
+      Caps::new(DISCO_NODE, COMPUTED_CAPS_HASH.clone()).into(),
+      ECaps2::new(vec![ecaps2_hash]).into(),
+      Element::builder("stats-id", ns::DEFAULT_NS)
+        .append("gst-meet")
+        .build(),
+      Element::builder("jitsi_participant_codecType", ns::DEFAULT_NS)
+        .append(config.video_codec.as_str())
+        .build(),
+      Element::builder("audiomuted", ns::DEFAULT_NS)
+        .append("false")
+        .build(),
+      Element::builder("videomuted", ns::DEFAULT_NS)
+        .append("false")
+        .build(),
+      Element::builder("nick", "http://jabber.org/protocol/nick")
+        .append(config.nick.as_str())
+        .build(),
+    ];
+    if let Some(region) = &config.region {
+      presence.extend([
+        Element::builder("jitsi_participant_region", ns::DEFAULT_NS)
+          .append(region.as_str())
+          .build(),
+        Element::builder("region", "http://jitsi.org/jitsi-meet")
+          .attr("id", region)
+          .build(),
+      ]);
+    }
+    presence.extend(
+      config
+        .extra_muc_features
+        .iter()
+        .cloned()
+        .map(|var| Feature { var })
+        .map(|feature| feature.into()),
+    );
+
     let conference = Self {
       glib_main_context,
       jid: xmpp_connection
@@ -165,17 +205,19 @@ impl JitsiConference {
       jingle_session: Arc::new(Mutex::new(None)),
       inner: Arc::new(Mutex::new(JitsiConferenceInner {
         state: JitsiConferenceState::Discovering,
+        presence,
         participants: HashMap::new(),
         on_participant: None,
         on_participant_left: None,
         on_colibri_message: None,
         connected_tx: Some(tx),
       })),
+      tls_insecure: xmpp_connection.tls_insecure,
     };
 
     xmpp_connection.add_stanza_filter(conference.clone()).await;
 
-    let iq = Iq::from_set(generate_id(), conference_stanza).with_to(Jid::Full(focus));
+    let iq = Iq::from_set(generate_id(), conference_stanza).with_to(focus);
     xmpp_connection.tx.send(iq.into()).await?;
 
     rx.await?;
@@ -220,22 +262,27 @@ impl JitsiConference {
   }
 
   #[tracing::instrument(level = "debug", err)]
-  async fn send_presence(&self, payloads: Vec<Element>) -> Result<()> {
+  async fn send_presence(&self, payloads: &[Element]) -> Result<()> {
     let mut presence = Presence::new(presence::Type::None).with_to(self.jid_in_muc()?);
-    presence.payloads = payloads;
+    presence.payloads = payloads.to_owned();
     self.xmpp_tx.send(presence.into()).await?;
     Ok(())
   }
 
   #[tracing::instrument(level = "debug", err)]
   pub async fn set_muted(&self, media_type: MediaType, muted: bool) -> Result<()> {
-    self
-      .send_presence(vec![
-        Element::builder(media_type.jitsi_muted_presence_element_name(), "")
-          .append(muted.to_string())
-          .build(),
-      ])
-      .await
+    let mut locked_inner = self.inner.lock().await;
+    let element = Element::builder(
+      media_type.jitsi_muted_presence_element_name(),
+      ns::DEFAULT_NS,
+    )
+    .append(muted.to_string())
+    .build();
+    locked_inner
+      .presence
+      .retain(|el| el.name() != media_type.jitsi_muted_presence_element_name());
+    locked_inner.presence.push(element);
+    self.send_presence(&locked_inner.presence).await
   }
 
   pub async fn pipeline(&self) -> Result<gstreamer::Pipeline> {
@@ -340,14 +387,12 @@ impl JitsiConference {
         if let Err(e) = f(self.clone(), participant.clone()).await {
           warn!("on_participant failed: {:?}", e);
         }
-        else {
-          if let Ok(pipeline) = self.pipeline().await {
-            gstreamer::debug_bin_to_dot_file(
-              &pipeline,
-              gstreamer::DebugGraphDetails::ALL,
-              &format!("participant-added-{}", participant.muc_jid.resource),
-            );
-          }
+        else if let Ok(pipeline) = self.pipeline().await {
+          gstreamer::debug_bin_to_dot_file(
+            &pipeline,
+            gstreamer::DebugGraphDetails::ALL,
+            &format!("participant-added-{}", participant.muc_jid.resource),
+          );
         }
       }
     }
@@ -374,14 +419,12 @@ impl JitsiConference {
       if let Err(e) = f(self.clone(), participant.clone()).await {
         warn!("on_participant failed: {:?}", e);
       }
-      else {
-        if let Ok(pipeline) = self.pipeline().await {
-          gstreamer::debug_bin_to_dot_file(
-            &pipeline,
-            gstreamer::DebugGraphDetails::ALL,
-            &format!("participant-added-{}", participant.muc_jid.resource),
-          );
-        }
+      else if let Ok(pipeline) = self.pipeline().await {
+        gstreamer::debug_bin_to_dot_file(
+          &pipeline,
+          gstreamer::DebugGraphDetails::ALL,
+          &format!("participant-added-{}", participant.muc_jid.resource),
+        );
       }
     }
   }
@@ -408,13 +451,13 @@ impl StanzaFilter for JitsiConference {
   #[tracing::instrument(level = "trace")]
   fn filter(&self, element: &Element) -> bool {
     element.attr("from") == Some(self.config.focus.to_string().as_str())
-      && element.is("iq", "jabber:client")
+      && element.is("iq", ns::DEFAULT_NS)
       || element
         .attr("from")
         .and_then(|from| from.parse::<BareJid>().ok())
         .map(|jid| jid == self.config.muc)
         .unwrap_or_default()
-        && (element.is("presence", "jabber:client") || element.is("iq", "jabber:client"))
+        && (element.is("presence", ns::DEFAULT_NS) || element.is("iq", ns::DEFAULT_NS))
   }
 
   #[tracing::instrument(level = "trace", err)]
@@ -437,49 +480,22 @@ impl StanzaFilter for JitsiConference {
           bail!("focus IQ failed");
         };
 
-        let jitsi_disco_info = DiscoInfoResult {
-          node: Some("http://jitsi.org/jitsimeet".to_string()),
-          identities: vec![],
-          features: vec![],
-          extensions: vec![],
-        };
-
-        let jitsi_disco_hash =
-          ecaps2::hash_ecaps2(&ecaps2::compute_disco(&jitsi_disco_info)?, Algo::Sha_256)?;
-        let mut presence = vec![
-          Muc::new().into(),
-          ECaps2::new(vec![jitsi_disco_hash]).into(),
-          Element::builder("stats-id", "").append("gst-meet").build(),
-          Element::builder("jitsi_participant_codecType", "")
-            .append(self.config.video_codec.as_str())
-            .build(),
-          Element::builder("jitsi_participant_region", "")
-            .append(self.config.region.as_str())
-            .build(),
-          Element::builder("audiomuted", "").append("false").build(),
-          Element::builder("videomuted", "").append("false").build(),
-          Element::builder("nick", "http://jabber.org/protocol/nick")
-            .append(self.config.nick.as_str())
-            .build(),
-          Element::builder("region", "http://jitsi.org/jitsi-meet")
-            .attr("id", &self.config.region)
-            .build(),
-        ];
-        presence.extend(
-          self
-            .config
-            .extra_muc_features
-            .iter()
-            .map(|feature| Element::builder("feature", "").attr("var", feature).build()),
-        );
-        self.send_presence(presence).await?;
-        self.inner.lock().await.state = JoiningMuc;
+        let mut locked_inner = self.inner.lock().await;
+        self.send_presence(&locked_inner.presence).await?;
+        locked_inner.state = JoiningMuc;
       },
       JoiningMuc => {
         let presence = Presence::try_from(element)?;
-        if BareJid::from(presence.from.as_ref().unwrap().clone()) == self.config.muc {
-          debug!("Joined MUC: {}", self.config.muc);
-          self.inner.lock().await.state = Idle;
+        if let Some(payload) = presence
+          .payloads
+          .iter()
+          .find(|payload| payload.is("x", ns::MUC_USER))
+        {
+          let muc_user = MucUser::try_from(payload.clone())?;
+          if muc_user.status.contains(&MucStatus::SelfPresence) {
+            debug!("Joined MUC: {}", self.config.muc);
+            self.inner.lock().await.state = Idle;
+          }
         }
       },
       Idle => {
@@ -492,19 +508,43 @@ impl StanzaFilter for JitsiConference {
                   iq.from.as_ref().unwrap(),
                   query.node
                 );
-                if query.node.is_none() {
+                if let Some(node) = query.node {
+                  match node.splitn(2, '#').collect::<Vec<_>>().as_slice() {
+                    // TODO: also support ecaps2, as we send it in our presence.
+                    [uri, hash]
+                      if *uri == DISCO_NODE && *hash == COMPUTED_CAPS_HASH.to_base64() =>
+                    {
+                      let mut disco_info = DISCO_INFO.clone();
+                      disco_info.node = Some(node);
+                      let iq = Iq::from_result(iq.id, Some(disco_info))
+                        .with_from(Jid::Full(self.jid.clone()))
+                        .with_to(iq.from.unwrap());
+                      self.xmpp_tx.send(iq.into()).await?;
+                    },
+                    _ => {
+                      let error = StanzaError::new(
+                        ErrorType::Cancel,
+                        DefinedCondition::ItemNotFound,
+                        "en",
+                        format!("Unknown disco#info node: {}", node),
+                      );
+                      let iq = Iq::from_error(iq.id, error)
+                        .with_from(Jid::Full(self.jid.clone()))
+                        .with_to(iq.from.unwrap());
+                      self.xmpp_tx.send(iq.into()).await?;
+                    },
+                  }
+                }
+                else {
                   let iq = Iq::from_result(iq.id, Some(DISCO_INFO.clone()))
                     .with_from(Jid::Full(self.jid.clone()))
                     .with_to(iq.from.unwrap());
                   self.xmpp_tx.send(iq.into()).await?;
                 }
-                else {
-                  panic!("don't know how to handle disco info node: {:?}", query.node);
-                }
               }
             },
-            IqType::Set(element) => {
-              if let Ok(jingle) = Jingle::try_from(element) {
+            IqType::Set(element) => match Jingle::try_from(element) {
+              Ok(jingle) => {
                 if let Some(Jid::Full(from_jid)) = iq.from {
                   if jingle.action == Action::SessionInitiate {
                     if from_jid.resource == "focus" {
@@ -541,10 +581,8 @@ impl StanzaFilter for JitsiConference {
                 else {
                   debug!("Received Jingle IQ from invalid JID: {:?}", iq.from);
                 }
-              }
-              else {
-                debug!("Received non-Jingle IQ");
-              }
+              },
+              Err(e) => debug!("IQ did not successfully parse as Jingle: {:?}", e),
             },
             IqType::Result(_) => {
               if let Some(jingle_session) = self.jingle_session.lock().await.as_mut() {
@@ -557,15 +595,41 @@ impl StanzaFilter for JitsiConference {
 
                   if let Some(colibri_url) = colibri_url {
                     info!("Connecting Colibri WebSocket to {}", colibri_url);
-                    let colibri_channel = ColibriChannel::new(&colibri_url).await?;
+                    let colibri_channel =
+                      ColibriChannel::new(&colibri_url, self.tls_insecure).await?;
                     let (tx, rx) = mpsc::channel(8);
                     colibri_channel.subscribe(tx).await;
+                    let colibri_channel_ = colibri_channel.clone();
                     jingle_session.colibri_channel = Some(colibri_channel);
 
                     let self_ = self.clone();
                     tokio::spawn(async move {
                       let mut stream = ReceiverStream::new(rx);
                       while let Some(msg) = stream.next().await {
+                        // Some message types are handled internally rather than passed to the on_colibri_message handler.
+
+                        // End-to-end ping
+                        if let ColibriMessage::EndpointMessage { to, from, msg_payload } = &msg {
+                          if let Some(to) = to {
+                            let my_endpoint_id = self_.jid.to_string().split('-').next().unwrap().to_owned();
+                            if to == &my_endpoint_id {
+                              match serde_json::from_value::<JsonMessage>(msg_payload.clone()) {
+                                Ok(JsonMessage::E2ePingRequest { id }) => {
+                                  if let Err(e) = colibri_channel_.send(ColibriMessage::EndpointMessage {
+                                    from: Some(my_endpoint_id),
+                                    to: from.clone(),
+                                    msg_payload: serde_json::to_value(JsonMessage::E2ePingResponse { id }).unwrap(),
+                                  }).await {
+                                    warn!("failed to send e2e ping response: {:?}", e);
+                                  }
+                                  continue;
+                                },
+                                _ => {},
+                              }
+                            }
+                          }
+                        }
+
                         let locked_inner = self_.inner.lock().await;
                         if let Some(f) = &locked_inner.on_colibri_message {
                           if let Err(e) = f(self_.clone(), msg).await {
@@ -657,14 +721,14 @@ impl StanzaFilter for JitsiConference {
                         if let Err(e) = f(self.clone(), participant.clone()).await {
                           warn!("on_participant failed: {:?}", e);
                         }
-                        else {
-                          if let Some(jingle_session) = self.jingle_session.lock().await.as_ref() {
-                            gstreamer::debug_bin_to_dot_file(
-                              &jingle_session.pipeline(),
-                              gstreamer::DebugGraphDetails::ALL,
-                              &format!("participant-added-{}", participant.muc_jid.resource),
-                            );
-                          }
+                        else if let Some(jingle_session) =
+                          self.jingle_session.lock().await.as_ref()
+                        {
+                          gstreamer::debug_bin_to_dot_file(
+                            &jingle_session.pipeline(),
+                            gstreamer::DebugGraphDetails::ALL,
+                            &format!("participant-added-{}", participant.muc_jid.resource),
+                          );
                         }
                       }
                     }
